/*
 * Copyright (C) 2009-2013 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file prebuffer_filter.c Paraslash's prebuffering filter. */

#include <regex.h>

#include "para.h"
#include "prebuffer_filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "filter.h"
#include "string.h"
#include "error.h"

/** Data specific to the prebuffer filter. */
struct private_prebuffer_data {
	/** The configuration data for this instance of the filter. */
	struct prebuffer_filter_args_info *conf;
	/** Number of bytes prebuffered or -1 if no longer prebuffering. */
	int prebuffered;
	/** End of prebuffering period. */
	struct timeval barrier;
};

static void prebuffer_pre_select(struct sched *s, struct task *t)
{
	struct filter_node *fn = container_of(t, struct filter_node, task);
	struct btr_node *btrn = fn->btrn;
	size_t iqs = btr_get_input_queue_size(btrn);
	struct private_prebuffer_data *ppd = fn->private_data;
	struct prebuffer_filter_args_info *conf = ppd->conf;
	struct timeval diff;

	t->error = 0;
	if (iqs == 0)
		return;
	if (ppd->barrier.tv_sec == 0) {
		struct timeval tv;
		PARA_INFO_LOG("prebuffer period %dms\n",
			conf->duration_arg);
		ms2tv(conf->duration_arg, &tv);
		tv_add(&tv, now, &ppd->barrier);
	}
	if (tv_diff(&ppd->barrier, now, &diff) < 0)
		return sched_min_delay(s);
	sched_request_timeout(&diff, s);
}

static void prebuffer_close(struct filter_node *fn)
{
	free(fn->private_data);
}

static void prebuffer_post_select(__a_unused struct sched *s, struct task *t)
{
	struct filter_node *fn = container_of(t, struct filter_node, task);
	struct btr_node *btrn = fn->btrn;
	size_t iqs = btr_get_input_queue_size(btrn);
	struct private_prebuffer_data *ppd = fn->private_data;
	struct prebuffer_filter_args_info *conf = ppd->conf;

	t->error = 0;
	if (ppd->barrier.tv_sec == 0)
		return;
	if (tv_diff(now, &ppd->barrier, NULL) < 0)
		return;
	if (iqs < conf->size_arg)
		return;
	btr_splice_out_node(btrn);
	t->error = -E_PREBUFFER_SUCCESS;
}

static int prebuffer_parse_config(int argc, char **argv, void **config)
{
	struct prebuffer_filter_args_info *conf = para_calloc(sizeof(*conf));
	int ret = -E_PREBUFFER_SYNTAX;

	prebuffer_filter_cmdline_parser(argc, argv, conf);
	ret = -ERRNO_TO_PARA_ERROR(EINVAL);
	if (conf->duration_arg < 0)
		goto err;
	if (conf->size_arg < 0)
		goto err;
	PARA_NOTICE_LOG("prebuffering %ims, %i bytes\n", conf->duration_arg,
		conf->size_arg);
	*config = conf;
	return 1;
err:
	free(conf);
	return ret;
}

static void prebuffer_open(struct filter_node *fn)
{
	struct private_prebuffer_data *ppd = para_calloc(sizeof(*ppd));

	ppd->conf = fn->conf;
	fn->private_data = ppd;
}

static void prebuffer_free_config(void *conf)
{
	prebuffer_filter_cmdline_parser_free(conf);
}

/**
 * The init function of the prebuffer filter.
 *
 * \param f Pointer to the struct to initialize.
 */
void prebuffer_filter_init(struct filter *f)
{
	struct prebuffer_filter_args_info dummy;

	prebuffer_filter_cmdline_parser_init(&dummy);
	f->open = prebuffer_open;
	f->close = prebuffer_close;
	f->parse_config = prebuffer_parse_config;
	f->free_config = prebuffer_free_config;
	f->pre_select = prebuffer_pre_select;
	f->post_select = prebuffer_post_select;
	f->help = (struct ggo_help) {
		.short_help = prebuffer_filter_args_info_help,
		.detailed_help = prebuffer_filter_args_info_detailed_help
	};
}
